@@ -1,10 +1,9 @@
-#![allow(clippy::needless_return)]
-
 mod acl;
 mod client;
 mod error;
 mod proto;
 mod record;
+mod session;
 mod util;
 
 pub use self::acl::{Acl, AuthId, AuthUser, Permission};
