@@ -1,27 +1,32 @@
 use std::time::Duration;
 
+use futures::future;
 use pretty_assertions::assert_eq;
 use rand::distributions::Standard;
 use rand::{self, Rng};
 use testcontainers::clients::Cli as DockerCli;
-use testcontainers::images::generic::{GenericImage, WaitFor};
-use testcontainers::Docker;
+use testcontainers::core::{Healthcheck, WaitFor};
+use testcontainers::images::generic::GenericImage;
+use tokio::select;
 use zookeeper_client as zk;
 
 fn random_data() -> Vec<u8> {
     let rng = rand::thread_rng();
-    return rng.sample_iter(Standard).take(32).collect();
+    rng.sample_iter(Standard).take(32).collect()
 }
 
 fn zookeeper_image() -> GenericImage {
-    return GenericImage::new("zookeeper:3.7.0")
-        .with_wait_for(WaitFor::message_on_stdout("PrepRequestProcessor (sid:"));
+    let healthcheck = Healthcheck::default()
+        .with_cmd(["./bin/zkServer.sh", "status"].iter())
+        .with_interval(Duration::from_secs(6))
+        .with_retries(30);
+    GenericImage::new("zookeeper", "3.7.0").with_healthcheck(healthcheck).with_wait_for(WaitFor::Healthcheck)
 }
 
 async fn example() {
     let docker = DockerCli::default();
     let zookeeper = docker.run(zookeeper_image());
-    let zk_port = zookeeper.get_host_port(2181).unwrap();
+    let zk_port = zookeeper.get_host_port(2181);
 
     let cluster = format!("127.0.0.1:{}", zk_port);
 
@@ -66,14 +71,14 @@ async fn example() {
 
 #[tokio::test]
 async fn test_example() {
-    tokio::spawn(example()).await.unwrap()
+    tokio::spawn(async move { example().await }).await.unwrap()
 }
 
 #[tokio::test]
 async fn test_no_node() {
     let docker = DockerCli::default();
     let zookeeper = docker.run(zookeeper_image());
-    let zk_port = zookeeper.get_host_port(2181).unwrap();
+    let zk_port = zookeeper.get_host_port(2181);
 
     let cluster = format!("127.0.0.1:{}", zk_port);
 
@@ -97,7 +102,7 @@ async fn test_no_node() {
 async fn test_data_node() {
     let docker = DockerCli::default();
     let zookeeper = docker.run(zookeeper_image());
-    let zk_port = zookeeper.get_host_port(2181).unwrap();
+    let zk_port = zookeeper.get_host_port(2181);
 
     let cluster = format!("127.0.0.1:{}", zk_port);
     let client = zk::Client::connect(&cluster, Duration::from_secs(20)).await.unwrap();
@@ -121,7 +126,7 @@ async fn test_data_node() {
 async fn test_create_sequential() {
     let docker = DockerCli::default();
     let zookeeper = docker.run(zookeeper_image());
-    let zk_port = zookeeper.get_host_port(2181).unwrap();
+    let zk_port = zookeeper.get_host_port(2181);
 
     let cluster = format!("127.0.0.1:{}", zk_port);
     let client = zk::Client::connect(&cluster, Duration::from_secs(20)).await.unwrap();
@@ -144,7 +149,7 @@ async fn test_create_sequential() {
 async fn test_descendants_number() {
     let docker = DockerCli::default();
     let zookeeper = docker.run(zookeeper_image());
-    let zk_port = zookeeper.get_host_port(2181).unwrap();
+    let zk_port = zookeeper.get_host_port(2181);
 
     let cluster = format!("127.0.0.1:{}", zk_port);
     let client = zk::Client::connect(&cluster, Duration::from_secs(20)).await.unwrap();
@@ -197,7 +202,7 @@ where
 async fn test_ephemerals() {
     let docker = DockerCli::default();
     let zookeeper = docker.run(zookeeper_image());
-    let zk_port = zookeeper.get_host_port(2181).unwrap();
+    let zk_port = zookeeper.get_host_port(2181);
 
     let cluster = format!("127.0.0.1:{}", zk_port);
     let client = zk::Client::connect(&cluster, Duration::from_secs(20)).await.unwrap();
@@ -263,7 +268,7 @@ async fn test_ephemerals() {
 async fn test_chroot() {
     let docker = DockerCli::default();
     let zookeeper = docker.run(zookeeper_image());
-    let zk_port = zookeeper.get_host_port(2181).unwrap();
+    let zk_port = zookeeper.get_host_port(2181);
 
     let cluster = format!("127.0.0.1:{}", zk_port);
     let client = zk::Client::connect(&cluster, Duration::from_secs(20)).await.unwrap();
@@ -304,7 +309,7 @@ async fn test_chroot() {
 async fn test_auth() {
     let docker = DockerCli::default();
     let zookeeper = docker.run(zookeeper_image());
-    let zk_port = zookeeper.get_host_port(2181).unwrap();
+    let zk_port = zookeeper.get_host_port(2181);
 
     let cluster = format!("127.0.0.1:{}", zk_port);
     let client = zk::Client::connect(&cluster, Duration::from_secs(20)).await.unwrap();
@@ -333,7 +338,7 @@ async fn test_auth() {
 async fn test_delete() {
     let docker = DockerCli::default();
     let zookeeper = docker.run(zookeeper_image());
-    let zk_port = zookeeper.get_host_port(2181).unwrap();
+    let zk_port = zookeeper.get_host_port(2181);
 
     let cluster = format!("127.0.0.1:{}", zk_port);
     let client = zk::Client::connect(&cluster, Duration::from_secs(20)).await.unwrap();
@@ -356,7 +361,7 @@ async fn test_delete() {
 async fn test_oneshot_watcher() {
     let docker = DockerCli::default();
     let zookeeper = docker.run(zookeeper_image());
-    let zk_port = zookeeper.get_host_port(2181).unwrap();
+    let zk_port = zookeeper.get_host_port(2181);
 
     let cluster = format!("127.0.0.1:{}", zk_port);
     let client = zk::Client::connect(&cluster, Duration::from_secs(20)).await.unwrap();
@@ -364,8 +369,15 @@ async fn test_oneshot_watcher() {
     let path = "/abc";
     let child_path = "/abc/efg";
 
+    // Drop or remove last watchers.
+    let (_, drop_watcher) = client.check_and_watch_stat(path).await.unwrap();
+    drop(drop_watcher);
+    let (_, remove_watcher) = client.check_and_watch_stat(child_path).await.unwrap();
+    remove_watcher.remove().await.unwrap();
+
     // Stat watcher for node creation.
     let (stat, stat_watcher) = client.check_and_watch_stat(path).await.unwrap();
+    let (_, child_stat_watcher) = client.check_and_watch_stat(child_path).await.unwrap();
     assert_eq!(stat, None);
     let create_options = zk::CreateOptions::new(zk::CreateMode::Persistent, zk::Acl::anyone_all());
     let (stat, _) = client.create(path, Default::default(), &create_options).await.unwrap();
@@ -380,10 +392,36 @@ async fn test_oneshot_watcher() {
     let (get_data, get_stat, get_watcher) = client.get_and_watch_data(path).await.unwrap();
     let (_, get_children_stat, get_children_watcher) = client.get_and_watch_children(path).await.unwrap();
     let (_, list_children_watcher) = client.list_and_watch_children(path).await.unwrap();
+    let (_, stat_watcher_same) = client.check_and_watch_stat(path).await.unwrap();
+    let (_, _, get_watcher_same) = client.get_and_watch_data(path).await.unwrap();
 
     assert_eq!((vec![], stat), (get_data, get_stat));
     assert_eq!(stat, get_children_stat);
 
+    let (_, stat_receiving_watcher) = client.check_and_watch_stat(path).await.unwrap();
+    let (_, _, get_receiving_watcher) = client.get_and_watch_data(path).await.unwrap();
+    let (_, stat_received_watcher) = client.check_and_watch_stat(path).await.unwrap();
+    let (_, _, get_received_watcher) = client.get_and_watch_data(path).await.unwrap();
+
+    // Drop or remove watchers before event.
+    let (_, drop_watcher1) = client.check_and_watch_stat(path).await.unwrap();
+    let (_, _, drop_watcher2) = client.get_and_watch_data(path).await.unwrap();
+    let (_, _, drop_watcher3) = client.get_and_watch_children(path).await.unwrap();
+    let (_, drop_watcher4) = client.list_and_watch_children(path).await.unwrap();
+
+    let (_, remove_watcher1) = client.check_and_watch_stat(path).await.unwrap();
+    let (_, _, remove_watcher2) = client.get_and_watch_data(path).await.unwrap();
+    let (_, _, remove_watcher3) = client.get_and_watch_children(path).await.unwrap();
+    let (_, remove_watcher4) = client.list_and_watch_children(path).await.unwrap();
+    drop(drop_watcher1);
+    drop(drop_watcher2);
+    drop(drop_watcher3);
+    drop(drop_watcher4);
+    remove_watcher1.remove().await.unwrap();
+    remove_watcher2.remove().await.unwrap();
+    remove_watcher3.remove().await.unwrap();
+    remove_watcher4.remove().await.unwrap();
+
     // Child creation.
     client.create(child_path, Default::default(), &create_options).await.unwrap();
 
@@ -392,11 +430,21 @@ async fn test_oneshot_watcher() {
     assert_eq!(child_event.path, path);
     assert_eq!(child_event, get_children_watcher.changed().await);
 
+    let event = child_stat_watcher.changed().await;
+    assert_eq!(event.event_type, zk::EventType::NodeCreated);
+    assert_eq!(event.path, child_path);
+
     eprintln!("child creation done");
 
     let (_, list_children_watcher) = client.list_and_watch_children(path).await.unwrap();
     let (_, _, get_children_watcher) = client.get_and_watch_children(path).await.unwrap();
 
+    let (_, drop_watcher1) = client.list_and_watch_children(path).await.unwrap();
+    let (_, _, drop_watcher2) = client.get_and_watch_children(path).await.unwrap();
+
+    let (_, remove_watcher1) = client.list_and_watch_children(path).await.unwrap();
+    let (_, _, remove_watcher2) = client.get_and_watch_children(path).await.unwrap();
+
     // Child deletion.
     client.delete(child_path, None).await.unwrap();
 
@@ -405,27 +453,77 @@ async fn test_oneshot_watcher() {
     assert_eq!(child_event.path, path);
     assert_eq!(child_event, get_children_watcher.changed().await);
 
+    // Drop or remove watchers after event delivered.
+    drop(drop_watcher1);
+    drop(drop_watcher2);
+    remove_watcher1.remove().await.unwrap();
+    remove_watcher2.remove().await.unwrap();
+
     let (_, list_children_watcher) = client.list_and_watch_children(path).await.unwrap();
     let (_, _, get_children_watcher) = client.get_and_watch_children(path).await.unwrap();
 
     eprintln!("child deletion done");
 
     // Node data change.
+
+    // Drop receiving watcher.
+    let mut get_receiving_watcher_future = get_receiving_watcher.changed();
+    select! {
+        biased;
+        _ = unsafe { Pin::new_unchecked(&mut get_receiving_watcher_future) } => {},
+        _ = future::ready(()) => {},
+    }
+    drop(get_receiving_watcher_future);
+    let mut stat_receiving_watcher_future = stat_receiving_watcher.changed();
+    select! {
+        biased;
+        _ = unsafe { Pin::new_unchecked(&mut stat_receiving_watcher_future) } => {},
+        _ = future::ready(()) => {},
+    }
+    drop(stat_receiving_watcher_future);
+
     client.set_data(path, Default::default(), None).await.unwrap();
 
+    // Drop probably received watcher.
+    let mut get_received_watcher_future = get_received_watcher.changed();
+    select! {
+        biased;
+        _ = unsafe { Pin::new_unchecked(&mut get_received_watcher_future) } => {},
+        _ = future::ready(()) => {},
+    }
+    drop(get_received_watcher_future);
+
     let node_event = get_watcher.changed().await;
     assert_eq!(node_event.event_type, zk::EventType::NodeDataChanged);
     assert_eq!(node_event.path, path);
     assert_eq!(node_event, stat_watcher.changed().await);
+    assert_eq!(node_event, get_watcher_same.changed().await);
+    assert_eq!(node_event, stat_watcher_same.changed().await);
+
+    // Drop received watcher.
+    let mut stat_received_watcher_future = stat_received_watcher.changed();
+    select! {
+        biased;
+        _ = unsafe { Pin::new_unchecked(&mut stat_received_watcher_future) } => {},
+        _ = future::ready(()) => {},
+    }
+    drop(stat_received_watcher_future);
 
     eprintln!("node data change done");
 
     let (_, _, get_watcher) = client.get_and_watch_data(path).await.unwrap();
     let (_, stat_watcher) = client.check_and_watch_stat(path).await.unwrap();
 
+    let (_, _, drop_watcher) = client.get_and_watch_data(path).await.unwrap();
+    let (_, remove_watcher) = client.check_and_watch_stat(path).await.unwrap();
+
     // Node deletion.
     client.delete(path, None).await.unwrap();
 
+    // Drop or remove watchers after event happened.
+    drop(drop_watcher);
+    remove_watcher.remove().await.unwrap();
+
     let node_event = get_watcher.changed().await;
     assert_eq!(node_event.event_type, zk::EventType::NodeDeleted);
     assert_eq!(node_event.path, path);
@@ -436,11 +534,51 @@ async fn test_oneshot_watcher() {
     eprintln!("node deletion done");
 }
 
+#[tokio::test]
+async fn test_persistent_watcher_passive_remove() {
+    let docker = DockerCli::default();
+    let zookeeper = docker.run(zookeeper_image());
+    let zk_port = zookeeper.get_host_port(2181);
+
+    let cluster = format!("127.0.0.1:{}", zk_port);
+
+    let client = zk::Client::connect(&cluster, Duration::from_secs(10)).await.unwrap();
+    let create_options = zk::CreateOptions::new(zk::CreateMode::Persistent, zk::Acl::anyone_all());
+
+    let path = "/abc";
+    let child_path = "/abc/efg";
+    let grandchild_path = "/abc/efg/123";
+
+    client.create(path, Default::default(), &create_options).await.unwrap();
+    client.create(child_path, Default::default(), &create_options).await.unwrap();
+    client.create(grandchild_path, Default::default(), &create_options).await.unwrap();
+    let (_, root_child_watcher) = client.list_and_watch_children("/").await.unwrap();
+
+    // Two watchers on same path, drop persistent watcher has no effect.
+    let (_, children_watcher) = client.list_and_watch_children(child_path).await.unwrap();
+    let persistent_watcher = client.watch(child_path, zk::AddWatchMode::Persistent).await.unwrap();
+    drop(persistent_watcher);
+
+    client.delete(grandchild_path, None).await.unwrap();
+    let child_event = children_watcher.changed().await;
+    assert_eq!(child_event.event_type, zk::EventType::NodeChildrenChanged);
+    assert_eq!(child_event.path, child_path);
+
+    // Now, no watcher remains, this deletion will detect dangling persistent watcher.
+    client.delete(child_path, None).await.unwrap();
+
+    // No other watchers should be affected.
+    client.delete(path, None).await.unwrap();
+    let child_event = root_child_watcher.changed().await;
+    assert_eq!(child_event.event_type, zk::EventType::NodeChildrenChanged);
+    assert_eq!(child_event.path, "/");
+}
+
 #[tokio::test]
 async fn test_persistent_watcher() {
     let docker = DockerCli::default();
     let zookeeper = docker.run(zookeeper_image());
-    let zk_port = zookeeper.get_host_port(2181).unwrap();
+    let zk_port = zookeeper.get_host_port(2181);
 
     let cluster = format!("127.0.0.1:{}", zk_port);
 
@@ -452,10 +590,41 @@ async fn test_persistent_watcher() {
     let grandchild_path = "/abc/efg/123";
     let unrelated_path = "/xyz";
 
+    // Remove last watchers.
+    let root_recursive_watcher = client.watch("/", zk::AddWatchMode::PersistentRecursive).await.unwrap();
+    let path_persistent_watcher = client.watch(path, zk::AddWatchMode::PersistentRecursive).await.unwrap();
+    drop(root_recursive_watcher);
+    path_persistent_watcher.remove().await.unwrap();
+    tokio::time::sleep(Duration::from_millis(10)).await;
+
     let mut root_recursive_watcher = client.watch("/", zk::AddWatchMode::PersistentRecursive).await.unwrap();
     let mut path_recursive_watcher = client.watch(path, zk::AddWatchMode::PersistentRecursive).await.unwrap();
     let mut path_persistent_watcher = client.watch(path, zk::AddWatchMode::Persistent).await.unwrap();
 
+    let root_recursive_watcher1 = client.watch("/", zk::AddWatchMode::PersistentRecursive).await.unwrap();
+    let path_persistent_watcher1 = client.watch(path, zk::AddWatchMode::Persistent).await.unwrap();
+    let root_recursive_watcher2 = client.watch("/", zk::AddWatchMode::PersistentRecursive).await.unwrap();
+    let path_persistent_watcher2 = client.watch(path, zk::AddWatchMode::Persistent).await.unwrap();
+    let mut root_recursive_watcher3 = client.watch("/", zk::AddWatchMode::PersistentRecursive).await.unwrap();
+    let path_persistent_watcher3 = client.watch(path, zk::AddWatchMode::Persistent).await.unwrap();
+
+    // Remove shared watch before events should have not effect on existing watchers.
+    drop(root_recursive_watcher1);
+    path_persistent_watcher1.remove().await.unwrap();
+
+    let mut root_recursive_watcher3_changed = root_recursive_watcher3.changed();
+    select! {
+        biased;
+        _ = unsafe { Pin::new_unchecked(&mut root_recursive_watcher3_changed) } => {},
+        _ = future::ready(()) => {},
+    }
+    let mut path_persistent_watcher3_remove = path_persistent_watcher3.remove();
+    select! {
+        biased;
+        _ = unsafe { Pin::new_unchecked(&mut path_persistent_watcher3_remove) } => {},
+        _ = future::ready(()) => {},
+    }
+
     // Node creation.
     client.create(path, Default::default(), &create_options).await.unwrap();
     let event = root_recursive_watcher.changed().await;
@@ -475,6 +644,10 @@ async fn test_persistent_watcher() {
     assert_eq!(child_event.event_type, zk::EventType::NodeChildrenChanged);
     assert_eq!(child_event.path, path);
 
+    // Remove shared watch after events should have not effect on existing watchers.
+    drop(root_recursive_watcher2);
+    path_persistent_watcher2.remove().await.unwrap();
+
     // Grandchild node creation.
     client.create(grandchild_path, Default::default(), &create_options).await.unwrap();
     let event = root_recursive_watcher.changed().await;
@@ -506,6 +679,10 @@ async fn test_persistent_watcher() {
     assert_eq!(event.event_type, zk::EventType::NodeCreated);
     assert_eq!(event.path, unrelated_path);
 
+    // Remove shared watch after events should have not effect on existing watchers.
+    drop(root_recursive_watcher3_changed);
+    drop(path_persistent_watcher3_remove);
+
     // Node deletion.
     client.delete(path, None).await.unwrap();
     let event = root_recursive_watcher.changed().await;
@@ -527,7 +704,7 @@ async fn test_persistent_watcher() {
 async fn test_session_event() {
     let docker = DockerCli::default();
     let zookeeper = docker.run(zookeeper_image());
-    let zk_port = zookeeper.get_host_port(2181).unwrap();
+    let zk_port = zookeeper.get_host_port(2181);
 
     let cluster = format!("127.0.0.1:{}", zk_port);
 
